package com.earldouglas.seekwell

object `package` {

  import java.sql._

  trait ColumnType[A]

  case class StringCT(columnName: String) extends ColumnType[String]
  def string(columnName: String): StringCT = StringCT(columnName)

  case class IntCT(columnName: String) extends ColumnType[Int]
  def int(columnName: String): IntCT = IntCT(columnName)

  case class TimestampCT(columnName: String) extends ColumnType[Timestamp]
  def timestamp(columnName: String): TimestampCT = TimestampCT(columnName)

  implicit def richPreparedStatment(stmt: java.sql.PreparedStatement): RichPreparedStatement = new RichPreparedStatement(stmt)

  class RichPreparedStatement(stmt: java.sql.PreparedStatement) {
    private val index = new java.util.concurrent.atomic.AtomicInteger()
    def extract[A](a: ColumnType[A]) = new Extractor1[A](stmt, a)
    def extract[A,B](a: ColumnType[A], b: ColumnType[B]) = new Extractor2[A, B](stmt, a, b)
    def extract[A,B,C](a: ColumnType[A], b: ColumnType[B], c: ColumnType[C]) = new Extractor3[A,B,C](stmt, a, b, c)
    def bind(value: String) = {
      stmt.setString(index.incrementAndGet(), value)
      this
    }
    def bind(value: Int) = {
      stmt.setInt(index.incrementAndGet(), value)
      this
    }
  }

  trait Extractor {
    def extract[C](rs: ResultSet, columnType: ColumnType[C]): C = columnType match {
      case StringCT(columnName)    => rs.getString(columnName)
      case IntCT(columnName)       => rs.getInt(columnName)
      case TimestampCT(columnName) => rs.getTimestamp(columnName)
    }
  }

  class Extractor1[A](stmt: PreparedStatement, a: ColumnType[A]) extends Extractor {
    def iterator(): Iterator[A] = {
      val rs: ResultSet = stmt.executeQuery()
      new RSIterator[A](stmt, rs, rs => extract[A](rs, a))
    }
  }

  class Extractor2[A, B](stmt: PreparedStatement, a: ColumnType[A], b: ColumnType[B]) extends Extractor {
    def iterator(): Iterator[Tuple2[A, B]] = {
      val rs: ResultSet = stmt.executeQuery()
      new RSIterator[Tuple2[A, B]](stmt, rs, rs => (extract[A](rs, a), extract[B](rs, b)))
    }
  }

  class Extractor3[A, B, C](stmt: PreparedStatement, a: ColumnType[A], b: ColumnType[B], c: ColumnType[C]) extends Extractor {
    def iterator(): Iterator[Tuple3[A, B, C]] = {
      val rs: ResultSet = stmt.executeQuery()
      new RSIterator[Tuple3[A, B, C]](stmt, rs, rs => (extract[A](rs, a), extract[B](rs, b), extract[C](rs, c)))
    }
  }

  class RSIterator[A](stmt: PreparedStatement, rs: ResultSet, f: ResultSet => A) extends Iterator[A] {
    private var moreResults: Boolean = rs.next()
    def hasNext(): Boolean = synchronized(moreResults)
    def next(): A = synchronized {
      val a = f(rs)
      if (hasNext()) {
        moreResults = rs.next()
      } else {
        rs.close()
        stmt.close()
      }
      a
    }
    def remove(): Unit = throw new java.lang.UnsupportedOperationException()
  }
}
